@@ -17,7 +17,7 @@ import signal
 from enum import Enum
 from numerous_api_client.headers.ValidationInterceptor import ValidationInterceptor
 import traceback
-
+from uuid import uuid4
 
 class JobStatus(Enum):
     ready=0
@@ -171,7 +171,13 @@ class NumerousClient:
         self._scenario = get_env(scenario, 'NUMEROUS_SCENARIO')
         self._token = get_env(token, 'NUMEROUS_API_TOKEN')
 
-        self._spm_job_id = get_env(job_id, 'SPM_JOB_ID')
+        self._job_id = get_env(job_id, 'JOB_ID')
+        self._execution_id = get_env(None, 'NUMEROUS_EXECUTION_ID', 'not_found')
+        if self._execution_id == 'not_found' or self._execution_id is None:
+            self._execution_id = str(uuid4())
+            log.debug(f'No execution id found - generating new: {self._execution_id}')
+
+        self._complete = True
 
         channel = self._init_channel(server=server, port=port)
         stub = spm_pb2_grpc.SPMStub(channel)
@@ -180,6 +186,10 @@ class NumerousClient:
         self._job_manager = spm_pb2_grpc.JobManagerStub(channel)
 
         self._base_client = stub
+
+        # Register execution
+        self._base_client.RegisterExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._job_id, execution_id=self._execution_id))
+
         self._terminated = False
         if not no_log:
             self._init_logger()
@@ -200,9 +210,12 @@ class NumerousClient:
         # Uncomment to enable timed probe feature...
         self.probe_doc_id = None
 
-        #self._listen = threading.Thread(target=self._listen_terminate)
-        #self._listen.setDaemon(True)
-        #self._listen.start()
+        self._listen = threading.Thread(target=self._listen_terminate)
+        self._listen.setDaemon(True)
+        self._listen.start()
+
+    def complete_execution(self):
+        self._base_client.CompleteExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._job_id, execution_id=self._execution_id))
 
     def _init_channel(self, server, port):
         secure_channel = os.getenv('SECURE_CHANNEL', False)
@@ -233,10 +246,16 @@ class NumerousClient:
         self.log_listener.start()
 
 
-    def _listen_terminate(self, scenario=None):
+    def _listen_terminate(self):
 
-       for doc in self._base_client.ListenScenario(spm_pb2.Scenario(project=self._project, scenario=self._scenario)):
-            pass
+       for channel, message in self.subscribe_messages([".".join(['COMMAND', self._project, self._scenario, self._job_id])]):
+            log.debug('Received message: '+str(message)+'\n on channel: '+str(channel))
+            if 'command' in message:
+                command = message['command']
+                if command == 'terminate':
+                    log.warning('Received Termination Command')
+                    signal.raise_signal(signal.SIGTERM)
+                    break
 
     def get_scenario_document(self, scenario=None):
         if scenario is None:
@@ -250,7 +269,7 @@ class NumerousClient:
         if scenario_doc is None:
             scenario_doc, files = self.get_scenario_document()
 
-        return scenario_doc['jobs'][self._spm_job_id]
+        return scenario_doc['jobs'][self._job_id]
 
 
     def get_group_document(self, group):
@@ -324,7 +343,7 @@ class NumerousClient:
 
             self.last_progress=tic
             self._base_client.SetScenarioProgress(spm_pb2.ScenarioProgress(
-                project=self._project, scenario=scenario, spm_job_id=self._spm_job_id,
+                project=self._project, scenario=scenario, spm_job_id=self._job_id,
                 message=message, status=status, clean=clean, progress=progress
             ))
 
@@ -352,7 +371,7 @@ class NumerousClient:
         if scenario is None:
             scenario = self._scenario
 
-        self._base_client.PushScenarioError(spm_pb2.ScenarioError(project=self._project, scenario=scenario, error=error, spm_job_id=self._spm_job_id))
+        self._base_client.PushScenarioError(spm_pb2.ScenarioError(project=self._project, scenario=scenario, error=error, spm_job_id=self._job_id))
 
     def push_scenario_log(self, message, initialize=False, scenario=None):
         if scenario is None:
@@ -360,7 +379,7 @@ class NumerousClient:
 
         self._base_client.PushScenarioLogEntry(spm_pb2.ScenarioLogEntry(
             project=self._project, scenario=scenario, message=message,
-            spm_job_id=self._spm_job_id, initialize=initialize
+            spm_job_id=self._job_id, initialize=initialize
         ))
 
     def delete_scenario(self, scenario=None):
@@ -738,25 +757,6 @@ class NumerousClient:
                 yield r
             return
 
-    # TODO: Remove? The client should not be allowed to stop/start instances of clients!
-    def start_image(self, scenario, image, name, namespace, spm_job_id, google_project='simulationwidgets'):
-        deployed_image = self._job_manager.StartImage(spm_pb2.ImageInformation(
-            project_id=self._project, scenario_id=scenario,
-            google_project=google_project, image=image,
-            deployment_name=name, namespace=namespace,
-            spm_job_id=spm_job_id
-        ))
-
-        return {'kubernetes_id': deployed_image.kubernetes_id, 'namespace': deployed_image.namespace}
-
-    # TODO: Remove? The client should not be allowed to stop/start instances of clients!
-    def stop_image(self, name, namespace):
-        stopped_image = self._job_manager.StopDeployment(spm_pb2.Deployment(
-            kubernetes_id=name, namespace=namespace
-        ))
-
-        return {'kubernetes_id': stopped_image.kubernetes_id, 'namespace': stopped_image.namespace}
-
     def publish_message(self, message):
         channel = ".".join(['EVENT', self._project, self._scenario])
 
@@ -765,12 +765,14 @@ class NumerousClient:
     #    self._base_client.SubscribeForUpdates(spm_pb2.)
     def subscribe_messages(self, channel_patterns):
         for message in self._base_client.SubscribeForUpdates(spm_pb2.Subscription(channel_patterns=channel_patterns)):
-            yield message.channel, message.message
+            yield message.channel, json.loads(message.message)
 
 
     def close(self):
         if hasattr(self,'log_listener'):
             self.log_listener.stop()
+        if self._complete:
+            self.complete_execution()
 
 
 @contextmanager
@@ -947,9 +949,39 @@ class NumerousAdminClient(NumerousClient):
 
         self._base_client = stub
         self._terminated = False
+        self._complete = False
 
         log.info('init admin client')
 
+    # TODO: Remove? The client should not be allowed to stop/start instances of clients!
+    def start_image(self, scenario, image, name, namespace, spm_job_id, google_project='simulationwidgets'):
+        deployed_image = self._job_manager.StartImage(spm_pb2.ImageInformation(
+            project_id=self._project, scenario_id=scenario,
+            google_project=google_project, image=image,
+            deployment_name=name, namespace=namespace,
+            spm_job_id=spm_job_id
+        ))
+
+        return {'kubernetes_id': deployed_image.kubernetes_id, 'namespace': deployed_image.namespace}
+
+    def launch_job(self, project, scenario, job):
+        self._job_manager.StartJob(spm_pb2.Job(
+            project_id=project, scenario_id=scenario, job_id=job
+        ))
+
+    def terminate_job(self, project, scenario, job):
+        terminated_job = self._job_manager.TerminateJob(spm_pb2.Job(
+            project_id=project, scenario_id=scenario, job_id=job
+        ))
+
+    def reset_job(self, project, scenario, job):
+        reset_job = self._job_manager.ResetJob(spm_pb2.Job(
+            project_id=project, scenario_id=scenario, job_id=job
+        ))
+
+
+
+
 @contextmanager
 def open_admin_client() -> NumerousAdminClient:
     numerous_client_ = NumerousAdminClient()
